@@ -1,6 +1,10 @@
 /* lkcd_common.c - core analysis suite
  *
  * Copyright (C) 1999, 2000, 2001, 2002 Mission Critical Linux, Inc.
+ * Copyright (C) 2002 Silicon Graphics, Inc. 
+ * Copyright (C) 2002 Free Software Foundation, Inc.
+ * Copyright (C) 2002, 2003, 2004 David Anderson
+ * Copyright (C) 2002, 2003, 2004 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,12 +22,13 @@
  * 02/29/00, 2.2    Bug fixes, new commands, options
  * 04/11/00, 2.3    Bug fixes, new command, options, initial PowerPC framework
  * 04/12/00  ---    Transition to BitKeeper version control
+ * 04/17/02  ---    Integration to LKCD and non-monotonically increasing dumps
  * 
  * BitKeeper ID: @(#)sgi_lkcd.c 1.3
  *
  * 09/28/00  ---    Transition to CVS version control
  *
- * CVS: $Revision: 1.3 $ $Date: 2002/01/30 19:28:36 $
+ * CVS: $Revision: 1.15 $ $Date: 2004/05/24 14:59:08 $
  */
 
 /*
@@ -68,7 +73,7 @@ struct lkcd_environment *lkcd = &lkcd_environment;
  *  Each version has its own dump initialization.
  */
 int
-lkcd_dump_init(FILE *fp, int fd)
+lkcd_dump_init(FILE *fp, int fd, char *dumpfile)
 {
 	switch (lkcd->version)
 	{
@@ -80,8 +85,15 @@ lkcd_dump_init(FILE *fp, int fd)
 		return(lkcd_dump_init_v2_v3(fp, fd));
 
         case LKCD_DUMP_V5:
+        case LKCD_DUMP_V6:
 		return(lkcd_dump_init_v5(fp, fd));
 
+        case LKCD_DUMP_V7:
+		return(lkcd_dump_init_v7(fp, fd, dumpfile));
+
+        case LKCD_DUMP_V8:
+		return(lkcd_dump_init_v8(fp, fd, dumpfile));
+
 	default:
 		return FALSE;
 	}
@@ -125,7 +137,7 @@ set_remote_lkcd_panic_data(ulong task, char *buf)
 		if (!(lkcd->panic_string = (char *)malloc(strlen(buf)+1))) {
 			fprintf(stderr, 
 			    "cannot malloc space for panic message!\n");
-			exit(1);
+			clean_exit(1);
 		}
 		strcpy(lkcd->panic_string, buf);
 	}
@@ -181,9 +193,18 @@ is_lkcd_compressed_dump(char *s)
 		return TRUE;
 
 	case LKCD_DUMP_V5:
+	case LKCD_DUMP_V6:
 		lkcd->version = LKCD_DUMP_V5;
 		return TRUE;
 
+	case LKCD_DUMP_V7:
+		lkcd->version = LKCD_DUMP_V7;
+		return TRUE;
+
+	case LKCD_DUMP_V8:
+		lkcd->version = LKCD_DUMP_V8;
+		return TRUE;
+
 	default:
 		lkcd_print("unsupported LKCD dump version: %ld (%lx)\n", 
 			version & ~(LKCD_DUMP_MCLX_V0|LKCD_DUMP_MCLX_V1), 
@@ -212,6 +233,14 @@ dump_dump_page(char *s, void *dp)
         case LKCD_DUMP_V5:
                 dump_dump_page_v5(s, dp);
                 break;
+
+        case LKCD_DUMP_V7:
+                dump_dump_page_v7(s, dp);
+		break;
+
+        case LKCD_DUMP_V8:
+                dump_dump_page_v8(s, dp);
+		break;
         }
 }
 
@@ -263,6 +292,14 @@ dump_header_only:
         case LKCD_DUMP_V5:
                 dump_lkcd_environment_v5(LKCD_DUMP_HEADER_ONLY);
                 break;
+
+        case LKCD_DUMP_V7:
+                dump_lkcd_environment_v7(LKCD_DUMP_HEADER_ONLY);
+		break;
+
+        case LKCD_DUMP_V8:
+                dump_lkcd_environment_v8(LKCD_DUMP_HEADER_ONLY);
+		break;
         }
 
         if (arg == LKCD_DUMP_HEADER_ONLY)
@@ -283,6 +320,14 @@ dump_page_only:
         case LKCD_DUMP_V5:
                 dump_lkcd_environment_v5(LKCD_DUMP_PAGE_ONLY);
                 break;
+
+        case LKCD_DUMP_V7:
+                dump_lkcd_environment_v7(LKCD_DUMP_PAGE_ONLY);
+		break;
+
+        case LKCD_DUMP_V8:
+                dump_lkcd_environment_v8(LKCD_DUMP_PAGE_ONLY);
+		break;
         }
 	if (arg == LKCD_DUMP_PAGE_ONLY)
 		return;
@@ -480,10 +525,10 @@ lkcd_memory_dump(FILE *fp)
 	                lkcd_print("  [%2d]: ", i);
 	                wrap = 0;
 	                while (phe && LKCD_VALID_PAGE(phe->pg_flags)) {
-				sprintf(buf, BITS32() ? "%llx@" : "%lx@",
-					phe->pg_addr);
+				sprintf(buf, "%llx@", 
+					(ulonglong)phe->pg_addr);
 				sprintf(&buf[strlen(buf)],
-	                        	"%lx,", phe->pg_hdr_offset);
+	                        	"%llx,", (ulonglong)phe->pg_hdr_offset);
 				lkcd_print("%18s", buf);
 
 	                        phe = phe->next;
@@ -542,12 +587,42 @@ lkcd_memory_dump(FILE *fp)
 		lkcd_print("\n");
 	}
 
+	lkcd_print("  dumpfile_index: %s\n", lkcd->dumpfile_index);
+	lkcd_print("             ifd: %d\n", lkcd->ifd);
+        lkcd_print("    memory_pages: %ld\n", lkcd->memory_pages);
+        lkcd_print(" page_offset_max: %ld\n", lkcd->page_offset_max);
+        lkcd_print("  page_index_max: %ld\n", lkcd->page_index_max);
+        lkcd_print("    page_offsets: %lx\n", lkcd->page_offsets);
+
 	lkcd->fp = fpsave;
 
         return pages;
 
 }
 
+static void
+lkcd_speedo(void)
+{
+        static int i = 0;
+
+        switch (++i%4) {
+        case 0:
+                printf("|\b");
+                break;
+        case 1:
+                printf("\\\b");
+                break;
+        case 2:
+                printf("-\b");
+                break;
+        case 3:
+                printf("/\b");
+                break;
+        }
+	fflush(stdout);
+}
+
+
 /*
  *  The lkcd_lseek() routine does the bulk of the work setting things up 
  *  so that the subsequent lkcd_read() simply has to do a bcopy().
@@ -572,50 +647,139 @@ lkcd_memory_dump(FILE *fp)
  *  Lastly, cache_page() stores the requested page's data.
  */
 int
-lkcd_lseek(unsigned long paddr)
+lkcd_lseek(physaddr_t paddr)
 {
         long i;
+	int err;
         int eof;
         void *dp;
-        long page;
-	uint64_t physaddr;
+	int pass;
+        long page = 0;
+	physaddr_t physaddr;
+	int seeked_to_page = 0;
 
 	dp = lkcd->dump_page;
 
-	lkcd->curpos = paddr & (lkcd->page_size-1);
-	lkcd->curpaddr = paddr & ~(lkcd->page_size-1);
+	lkcd->curpos = paddr & ((physaddr_t)(lkcd->page_size-1));
+        lkcd->curpaddr = paddr & ~((physaddr_t)(lkcd->page_size-1));
 
 	if (page_is_cached()) 
 		return TRUE;
 
-	if (!page_is_hashed(&page) && !get_closest_page(&page))
-                return FALSE;
+	/* Faster than paging in lkcd->page_offsets[page] */
+	if(page_is_hashed(&page)) {
+		seeked_to_page = 1;
+	}
 
-        for (i = page, eof = FALSE; !eof; i++) {
+	/* 
+	 * All of the pages offsets should have been saved, 
+	 * so if we have plenty of swap space we simply page 
+	 * in the offset and seek.
+	 */
+	if (!seeked_to_page && lkcd->page_offsets != NULL) {
+		long page_index;
+		off_t page_offset;
+		off_t seek_offset;
 
-                switch (lkcd_load_dump_page_header(dp, i))
-                {
-                case LKCD_DUMPFILE_OK:
-			break;
-
-                case LKCD_DUMPFILE_END:
-                case LKCD_DUMPFILE_EOF:
-			eof = TRUE;
-                        continue;
-                }
-
-        	physaddr = lkcd->get_dp_flags() & 
-		    (LKCD_DUMP_MCLX_V0|LKCD_DUMP_MCLX_V1) ?
-                    (lkcd->get_dp_address() - lkcd->kvbase) << lkcd->page_shift:
-                    lkcd->get_dp_address() - lkcd->kvbase;
-
-		if (physaddr == lkcd->curpaddr) 
+		page_index = (lkcd->curpaddr >> lkcd->page_shift);
+		page_offset = lkcd->page_offsets[page_index];
+		if( page_offset > 0 ) {
+			seek_offset = lseek(lkcd->fd, page_offset, SEEK_SET);
+			if (seek_offset == page_offset) {
+				seeked_to_page = 1;
+				page = page_index;
+			}
+		}
+	}
+	if (seeked_to_page) {
+		err = lkcd_load_dump_page_header(dp, page);
+		if (err == LKCD_DUMPFILE_OK) {
 			return(cache_page());
+		}
+	}	
+		
 
-		if (physaddr > lkcd->curpaddr)
+	/*
+	 * We haven't seen this page yet or we didn't have
+	 * space for lkcd->page_offsets[].
+	 */
+	for (pass = 1; pass <= 3; pass++) {
+		switch (pass) {
+		case 1:
+			/* 
+			 * If we have lkcd->page_offsets,
+			 * then we know the offsets for
+			 * all offsets before our largest
+			 * seen. So we start there.
+			 */
+			if ( lkcd->page_offsets != NULL ) {
+				lseek(lkcd->fd, lkcd->page_offset_max, SEEK_SET);
+				page = lkcd->page_index_max;
+                        } else {
+                                continue;
+                        }
 			break;
 
-        	lseek(lkcd->fd, lkcd->get_dp_size(), SEEK_CUR);
+		case 2:
+			/* 
+			 * We don't have lkcd->page_offsets,
+			 * but we might get lucky by looking
+			 * for the page next to a neighboring
+			 * page.
+			 */
+			if (!get_closest_page(&page)) {
+	               		continue;
+			}
+			break;
+
+		case 3:
+			/* 
+			 * We don't have lkcd->page_offsets, so
+			 * we have to use brute force and search
+			 * from the beginning.
+			 */
+			lseek(lkcd->fd, LKCD_OFFSET_TO_FIRST_PAGE, SEEK_SET);
+			page = 0;
+			break;
+		}
+	
+	        for (i = page, eof = FALSE; !eof; i++) {
+
+			if( (i%128) == 0) {
+				lkcd_speedo();
+			}
+	
+	                switch (lkcd_load_dump_page_header(dp, i))
+	                {
+	                case LKCD_DUMPFILE_OK:
+				break;
+	
+	                case LKCD_DUMPFILE_END:
+	                case LKCD_DUMPFILE_EOF:
+				eof = TRUE;
+	                        continue;
+	                }
+	
+	        	physaddr = lkcd->get_dp_flags() & 
+			    (LKCD_DUMP_MCLX_V0|LKCD_DUMP_MCLX_V1) ?
+	                    (lkcd->get_dp_address() - lkcd->kvbase) << lkcd->page_shift:
+	                    lkcd->get_dp_address() - lkcd->kvbase;
+	
+			if (physaddr == lkcd->curpaddr) {
+				/* 
+				 * Update the mmap'd index file cache if it exists 
+				 * and return with the cached page.
+				 */
+                                if (lkcd->page_offsets != NULL) {
+                                        msync(&lkcd->page_offsets[0],
+                                              	(lkcd->page_index_max *
+						sizeof(off_t)), MS_ASYNC);
+                                }
+				return(cache_page());
+			}
+	
+	        	lseek(lkcd->fd, lkcd->get_dp_size(), SEEK_CUR);
+		}
 	}
 
 	return FALSE;
@@ -1004,7 +1168,7 @@ static int
 lkcd_uncompress_gzip(unsigned char *dest, ulong destlen, 
 	unsigned char *source, ulong sourcelen)
 {
-	ulong retlen;
+        ulong retlen = destlen;
 
 	switch (uncompress(dest, &retlen, source, sourcelen)) 
 	{
@@ -1067,8 +1231,9 @@ lkcd_load_dump_page_header(void *dp, ulong page)
 {
 	uint32_t dp_flags;
 	uint64_t dp_address, physaddr;
+	off_t page_offset;
 
-        lkcd->curhdroffs = lseek(lkcd->fd, 0, SEEK_CUR);
+        page_offset = lkcd->curhdroffs = lseek(lkcd->fd, 0, SEEK_CUR);
 
         if (read(lkcd->fd, dp, lkcd->page_header_size) != 
 	    lkcd->page_header_size) {
@@ -1087,13 +1252,54 @@ lkcd_load_dump_page_header(void *dp, ulong page)
 	if ((lkcd->flags & LKCD_VALID) && (page > lkcd->total_pages)) 
 		lkcd->total_pages = page;
 
-        if (page == 0)
+#ifdef X86
+	/*
+	 *  Ugly leftover from very early x86 LKCD versions which used 
+	 *  the kernel unity-mapped virtual address as the dp_address.
+	 */
+        if ((page == 0) && !(lkcd->flags & LKCD_VALID) && 
+	    (lkcd->version == LKCD_DUMP_V1) && 
+	    (dp_address == 0xc0000000)) 
         	lkcd->kvbase = dp_address;
+#endif
 
 	physaddr = dp_flags & (LKCD_DUMP_MCLX_V0|LKCD_DUMP_MCLX_V1) ?
 		(dp_address - lkcd->kvbase) << lkcd->page_shift : 
         	dp_address - lkcd->kvbase;
 
+	/*
+	 * Update page_offsets if possible and needed.
+	 */
+	if(lkcd->page_offsets != NULL) {
+		long page_index;
+
+		page_index = physaddr >> lkcd->page_shift;
+		if (page_index > lkcd->memory_pages) {
+			abort();
+		}
+		if ( lkcd->page_offsets[page_index] == 0 ) {
+			if (lkcd->page_offset_max > page_offset) {
+				/* 
+				 * Can't Happen - No holes allowed in page_offsets[]:
+				 *
+				 * It cost us a lot to seek and read the page headers.
+				 * Once we read it we never dispose of it. So if we
+				 * are looking for a page and we don't have the offset
+				 * for it then we haven't seen it and need to continue
+				 * reading the dump.
+				 */
+				abort();
+			}
+			lkcd->page_offsets[page_index] = page_offset;
+			lkcd->page_index_max = page_index;
+			lkcd->page_offset_max = page_offset;
+		} else {
+			if (lkcd->page_offsets[page_index] != lkcd->curhdroffs) {
+				abort();
+			}
+		}
+	}
+
         if (MEGABYTE_ALIGNED(physaddr) && !LKCD_BENCHMARKS_DONE()) 
                 set_mb_benchmark((ulong)(physaddr >> lkcd->page_shift));
 
